import pandas as pd
import numpy as np
from typing import List, Tuple, Dict

def compare_df(
    df1: pd.DataFrame,
    df2: pd.DataFrame,
    key: str,
    cols_df1: List[str],
    cols_df2: List[str]
) -> Tuple[pd.DataFrame, Dict[str, int]]:
    """
    Compares two DataFrames by joining on a key column and checks column-wise matches, considering type differences.
    Returns a binary DataFrame indicating matches (1) and mismatches (0), along with a dictionary containing mismatch counts.

    Parameters:
    -----------
    df1 : pd.DataFrame
        First DataFrame.
    df2 : pd.DataFrame
        Second DataFrame.
    key : str
        Key column name to join on.
    cols_df1 : List[str]
        List of column names in df1 to compare.
    cols_df2 : List[str]
        List of column names in df2 to compare.

    Returns:
    --------
    df_out : pd.DataFrame
        DataFrame with match indicators (1 for matches, 0 for mismatches).
    stats_dict : Dict[str, int]
        Dictionary with mismatch counts for each column comparison.

    Raises:
    -------
    ValueError:
        If column lists do not have the same length.
    """

    # Ensure column lists have the same length
    if len(cols_df1) != len(cols_df2):
        raise ValueError("Column lists must have the same length")

    # Select only necessary columns from both dataframes
    df1_filtered = df1[[key] + cols_df1].copy()
    df2_filtered = df2[[key] + cols_df2].copy()

    # Merge DataFrames on the key
    merged_df = pd.merge(df1_filtered, df2_filtered, on=key, how="inner")

    # Initialize output DataFrame and statistics dictionary
    df_out = pd.DataFrame(index=merged_df.index)
    stats_dict: Dict[str, int] = {}

    # Compare corresponding columns from both dataframes
    for col1, col2 in zip(cols_df1, cols_df2):
        col_df1 = merged_df[col1].apply(lambda x: np.nan if isinstance(x,str) and x.strip() == "" else x)  # Replace empty strings with NaN
        col_df2 = merged_df[col2].apply(lambda x: np.nan if isinstance(x,str) and x.strip() == "" else x)  # Replace empty strings with NaN

        # Check if values are both NaN (empty match)
        is_empty_match = col_df1.isna() & col_df2.isna()

        # Try numeric comparison
        try:
            col_df1, col_df2 = col_df1.astype(float), col_df2.astype(float)
            numeric_match = (col_df1 - col_df2).abs() < 1e-9
        except ValueError as e:
            raise ValueError(e)
            numeric_match = col_df1.astype(str) == col_df2.astype(str)  # Fallback for non-numeric data

        # Final match condition
        match_series = (numeric_match | is_empty_match).astype(int)

        # Store results
        df_out[f"{col1}_vs_{col2}"] = match_series
        stats_dict[f"{col1}_vs_{col2}"] = int((match_series == 0).sum())  # Count mismatches

    df_out[key] = merged_df[key]  # Keep the key column in the output DataFrame

    return df_out, stats_dict
